#!/usr/bin/env python

import skimage.restoration
from skimage import img_as_float
import numpy as np
from PIL import Image
import os
from tqdm import tqdm
import csv
import rawpy

imagesFolderPath = '/mnt/HDD0/raise'
imagesFolderPathFileName = imagesFolderPath.replace('/', '_')
denoiser = 'wavelet'
npArrayFilePath = f'mean_{imagesFolderPathFileName}_{denoiser}.npy'

denoise = getattr(skimage.restoration, f'denoise_{denoiser}')

mean = None
numberOfImagesInMean = 0

imagesFileNames = os.listdir(imagesFolderPath + '/png')

requiresRaiseFiltering = True
if requiresRaiseFiltering:
    files = {}

    with open('RAISE_all.csv') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in tqdm(list(reader), 'CSV parsing'):
            file = row['File'] + '.png'
            files[file] = row

    imagesFileNames = [imageFileName for imageFileName in tqdm(imagesFileNames, 'Filtering images') if files[imageFileName]['Device'] == 'Nikon D7000' and Image.open(f'{imagesFolderPath}/png/{imageFileName}').size == (4946, 3278)]
#imagesFileNames = [f'DSC0{imageIndex}.ARW' for imageIndex in range(2807, 2911)]

minColor = None
maxColor = None

def treatImage(imageFileName, computeExtremes = False):
    global mean, numberOfImagesInMean, minColor, maxColor
    imageFilePath = f'{imagesFolderPath}/nef/{imageFileName}'
    if imageFileName.endswith('.NEF') or imageFileName.endswith('.ARW'):
        with rawpy.imread(imageFilePath) as raw:
            colorDesc = raw.color_desc.decode('ascii')
            assert colorDesc == 'RGBG'
            assert np.array_equal(raw.raw_pattern, np.array([[0, 1], [3, 2]], dtype = np.uint8))
            rawImageVisible = raw.raw_image_visible.copy()
            greenRawImageVisible = rawImageVisible[1::2, ::2]
            if computeExtremes:
                greenRawImageVisibleMin = greenRawImageVisible.min()
                greenRawImageVisibleMax = greenRawImageVisible.max()
                if minColor is None or greenRawImageVisibleMin < minColor:
                    minColor = greenRawImageVisibleMin
                if maxColor is None or greenRawImageVisibleMax > maxColor:
                    maxColor = greenRawImageVisibleMax
                return
            imageNpArray = (greenRawImageVisible - minColor) / (maxColor - minColor)
            # Pay attention to range of values expected by the denoiser.
            # Indeed if provide the thousands valued raw image, then the denoiser only returns values between 0 and 1 and making the difference between both look pointless.
    else:
        imagePil = Image.open(imageFilePath)
        imageNpArray = img_as_float(np.array(imagePil))
    # Could make not `.nef` compatible.
    match denoiser:
        case 'wavelet':
            imageDenoisedNpArray = denoise(imageNpArray, rescale_sigma=True)
        case 'bilateral':
            imageDenoisedNpArray = denoise(imageNpArray, sigma_color=0.05, sigma_spatial=15)
        case 'tv_chambolle':
            imageDenoisedNpArray = denoise(imageNpArray, weight=0.2)
    imageNoiseNpArray = imageNpArray - imageDenoisedNpArray
    if mean is None:
        mean = imageNoiseNpArray
    else:
        mean = ((mean * numberOfImagesInMean) + imageNoiseNpArray) / (numberOfImagesInMean + 1)
    numberOfImagesInMean += 1

for imageFileName in tqdm(imagesFileNames, 'Computing extremes of images'):
    treatImage(imageFileName, computeExtremes = True)

print(f'{minColor=}')
print(f'{maxColor=}')

for imageFileName in tqdm(imagesFileNames, 'Denoising images'):
    treatImage(imageFileName)

with open(npArrayFilePath, 'wb') as f:
    np.save(f, mean)
